@@ -23,7 +23,13 @@ from pip_audit._dependency_source import (
 )
 from pip_audit._dependency_source.interface import DependencySourceError
 from pip_audit._fix import ResolvedFixVersion, SkippedFixVersion, resolve_fix_versions
-from pip_audit._format import ColumnsFormat, CycloneDxFormat, JsonFormat, VulnerabilityFormat
+from pip_audit._format import (
+    ColumnsFormat,
+    CycloneDxFormat,
+    JsonFormat,
+    MarkdownFormat,
+    VulnerabilityFormat,
+)
 from pip_audit._service import OsvService, PyPIService, VulnerabilityService
 from pip_audit._service.interface import ConnectionError as VulnServiceConnectionError
 from pip_audit._service.interface import ResolvedDependency, SkippedDependency
@@ -44,6 +50,7 @@ class OutputFormatChoice(str, enum.Enum):
     Json = "json"
     CycloneDxJson = "cyclonedx-json"
     CycloneDxXml = "cyclonedx-xml"
+    Markdown = "markdown"
 
     def to_format(self, output_desc: bool) -> VulnerabilityFormat:
         if self is OutputFormatChoice.Columns:
@@ -54,6 +61,8 @@ class OutputFormatChoice(str, enum.Enum):
             return CycloneDxFormat(inner_format=CycloneDxFormat.InnerFormat.Json)
         elif self is OutputFormatChoice.CycloneDxXml:
             return CycloneDxFormat(inner_format=CycloneDxFormat.InnerFormat.Xml)
+        elif self is OutputFormatChoice.Markdown:
+            return MarkdownFormat(output_desc)
         else:
             assert_never(self)
 
@@ -303,11 +312,13 @@ def _parse_args(parser: argparse.ArgumentParser) -> argparse.Namespace:
     return parser.parse_args()
 
 
-def _dep_source_from_project_path(project_path: Path, state: AuditState) -> DependencySource:
+def _dep_source_from_project_path(
+    project_path: Path, resolver: ResolveLibResolver, state: AuditState
+) -> DependencySource:
     # Check for a `pyproject.toml`
     pyproject_path = project_path / "pyproject.toml"
     if pyproject_path.is_file():
-        return PyProjectSource(pyproject_path, ResolveLibResolver(), state)
+        return PyProjectSource(pyproject_path, resolver, state)
 
     # TODO: Checks for setup.py and other project files will go here.
 
@@ -348,6 +359,12 @@ def audit() -> None:
     if args.require_hashes and args.no_deps:
         logger.warning("The --no-deps flag is redundant when used with --require-hashes")
 
+    if args.require_hashes and isinstance(service, OsvService):
+        logger.warning(
+            "The --require-hashes flag with --service osv only enforces hash presence NOT hash "
+            "validity. Use --service pypi to enforce hash validity."
+        )
+
     if args.no_deps:
         logger.warning(
             "--no-deps is supported, but users are encouraged to fully hash their "
@@ -384,7 +401,13 @@ def audit() -> None:
             # once PEP 660 is more widely supported: https://www.python.org/dev/peps/pep-0660/
 
             # Determine which kind of project file exists in the project path
-            source = _dep_source_from_project_path(args.project_path, state)
+            source = _dep_source_from_project_path(
+                args.project_path,
+                ResolveLibResolver(
+                    index_urls, args.timeout, args.cache_dir, args.skip_editable, state
+                ),
+                state,
+            )
         else:
             source = PipSource(
                 local=args.local, paths=args.paths, skip_editable=args.skip_editable, state=state
